# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""RIOT Driver for RIOT PAL
This module handles generic connection and IO to the make term.
"""


class RiotDriver:
    """Contains all reusable functions for connecting, sending and recieveing
    data.

    """

    def __init__(self):
        raise NotImplementedError()

    def close(self):
        """Close serial connection."""
        raise NotImplementedError()

    def read(self):
        """Read and decode data."""
        raise NotImplementedError()

    def write(self, data):
        """Tries write data."""
        raise NotImplementedError()



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Low Level Shell for RIOT PAL
This module handles functions for a low level shell interface.
"""
import logging
import errno
import os
from .base_device import BaseDevice


class LLShell(BaseDevice):
    """Handles basic functions and commands for memory map interface."""
    READ_REG_CMD = "rr"
    WRITE_REG_CMD = "wr"
    EXECUTE_CMD = "ex"
    RESET_CMD = "mcu_rst"
    SUCCESS = '0'
    RESET_SUCCESS = '\x000'
    RESULT_SUCCESS = 'Success'
    RESULT_ERROR = 'Error'
    RESULT_TIMEOUT = 'Timeout'

    @staticmethod
    def _try_parse_data(data, to_byte_array):
        if len(data) > 1:
            # response contains data
            try:
                if ((len(data[1]) - 2) <= 16) and not to_byte_array:
                    return int(data[1], 0)
                elif data[1].startswith('0x'):
                    data = bytearray.fromhex(data[1][2:])
                    return list(data[::-1])
                else:
                    return data[1:]
            except ValueError:
                return data[1:]

        return None

    @staticmethod
    def _error_msg(data):
        if data not in errno.errorcode:
            return "Unknown Error[{}]".format(data)
        s_errcode = errno.errorcode[data]
        s_errmsg = os.strerror(data)
        return "{}-{} [{}]".format(s_errcode, s_errmsg, data)

    def _populate_cmd_info(self, data, to_byte_array):
        cmd_info = {}
        try:
            # handle case if reset occurs and there is a 0 in the buffer
            if data[0] == self.RESET_SUCCESS:
                data[0] = self.SUCCESS
            if data[0] == self.SUCCESS:
                cmd_info['data'] = self._try_parse_data(data, to_byte_array)
                cmd_info['msg'] = "EOK-command success [0]"
                cmd_info['result'] = self.RESULT_SUCCESS
                logging.debug(self.RESULT_SUCCESS)
            else:
                # put error code in data
                cmd_info['data'] = int(data[0], 0)
                cmd_info['msg'] = self._error_msg(cmd_info['data'])
                cmd_info['result'] = self.RESULT_ERROR
                logging.debug(self.RESULT_ERROR)
                logging.debug(cmd_info['msg'])
        except (ValueError, TypeError, AttributeError) as exc:
            cmd_info['msg'] = "Unknown Error {}".format(exc)
            cmd_info['data'] = data[0]
            cmd_info['result'] = self.RESULT_ERROR
            logging.debug(self.RESULT_ERROR)
            logging.debug(exc)
        return cmd_info

    def send_cmd(self, send_cmd, to_byte_array=False):
        """Returns a dictionary with information from the event.
        Returns:
            dict:
            The return hold dict values in the following keys::
            msg - The message from the response, only used for information.
            cmd - The command sent, used to track what has occured.
            data - Parsed information of the data requested.
            result - Either success, error or timeout.
        """
        self._write(send_cmd)
        data = self._read()
        cmd_info = {'cmd': send_cmd}
        if data == "":
            cmd_info['msg'] = "Timeout occured"
            cmd_info['data'] = None
            cmd_info['result'] = self.RESULT_TIMEOUT
            logging.debug(self.RESULT_TIMEOUT)
        else:
            data = data.replace('\n', '')
            data = data.split(',')
            cmd_info.update(self._populate_cmd_info(data, to_byte_array))
        return cmd_info

    def read_bytes(self, index, size=1, to_byte_array=False):
        """Reads bytes in the register map."""
        logging.debug("FXN: read_bytes(%r,%r)", index, size)
        cmd = '{} {} {}'.format(self.READ_REG_CMD, index, size)
        return self.send_cmd(cmd, to_byte_array)

    def write_bytes(self, index, data, size=4):
        """Writes bytes in the register map."""
        logging.debug("FXN: write_bytes(%r,%r)", index, data)
        cmd = "{} {}".format(self.WRITE_REG_CMD, index)
        if isinstance(data, list):
            if isinstance(data[0], str):
                if data[0].startswith('0x'):
                    data = data[0].replace('0x', '')
                    data = bytes.fromhex(data)
            for i in range(0, len(data)):
                cmd += ' {}'.format(data[i])
        else:
            for i in range(0, size):
                cmd += ' {}'.format((data >> ((i) * 8)) & 0xFF)
        return self.send_cmd(cmd)

    def read_bits(self, index, offset, bit_amount):
        """Read specific bits in the register map."""
        bit_amount = int(bit_amount)
        offset = int(offset)
        index = int(index)
        logging.debug("FXN: read_bits(%r, %r, %r)", index, offset, bit_amount)
        bytes_to_read = int((bit_amount - 1 + offset)/8 + 1)
        bit_mask = (2 ** bit_amount) - 1
        cmd_info = self.read_bytes(index, bytes_to_read)
        if cmd_info['result'] == self.RESULT_SUCCESS:
            cmd_info['cmd'] += ', read_bits {} {} {}'.format(index, offset,
                                                             bit_amount)
            cmd_info['data'] = cmd_info['data'] >> offset
            cmd_info['data'] = cmd_info['data'] & bit_mask

        logging.debug("Bits: %r", cmd_info['data'])
        return cmd_info

    def write_bits(self, index, offset, bit_amount, data):
        """Modifies specific bits in the register map."""
        bit_amount = int(bit_amount)
        offset = int(offset)
        index = int(index)
        cmd_sent = ""
        logging.debug("FXN: write_bits"
                      "(%r, %r, %r, %r)", index, offset, bit_amount, data)
        bytes_to_read = int((bit_amount - 1 + offset)/8 + 1)
        cmd_info = self.read_bytes(index, bytes_to_read)
        if cmd_info['result'] != self.RESULT_SUCCESS:
            return cmd_info
        cmd_sent += cmd_info['cmd']
        bit_mask = int((2 ** bit_amount) - 1)
        bit_mask = bit_mask << offset
        cmd_info['data'] = cmd_info['data'] & (~bit_mask)
        shifted_data = cmd_info['data'] | ((data << offset) & bit_mask)
        cmd_info = self.write_bytes(index, shifted_data, bytes_to_read)
        cmd_sent += cmd_info['cmd']
        if cmd_info['result'] == self.RESULT_SUCCESS:
            cmd_sent += ',write_bits {} {} {} {}'.format(index, offset,
                                                         bit_amount, data)
        cmd_info['cmd'] = cmd_sent
        return cmd_info

    def execute_changes(self):
        """Executes device configuration changes."""
        logging.debug("FXN: execute_changes")
        return self.send_cmd(self.EXECUTE_CMD)

    def reset_mcu(self):
        """Resets the device."""
        logging.debug("FXN: reset_mcu")
        return self.send_cmd(self.RESET_CMD)



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Low Level Memory Map Interface for RIOT PAL
This module handles offset and sizes dictated a memory map.
"""
import logging
import csv
from .ll_shell import LLShell


class LLMemMapIf(LLShell):
    """Interface between ll_shell and mem_map."""

    def __init__(self, path, *args, **kwargs):
        self.cmd_list = LLMemMapIf.import_mm_from_csv(path)
        super().__init__(*args, **kwargs)

    @staticmethod
    def import_mm_from_csv(path):
        """Imports a memory map csv file."""
        cmd_list = {}
        with open(path) as csvfile:
            rows = list(csv.reader(csvfile, quotechar="'"))
            for row in range(1, len(rows)):
                cmd = dict(zip(rows[0], rows[row]))
                cmd_list[rows[row][rows[0].index('name')]] = cmd
                logging.debug("Imported command: %r", cmd)
        return cmd_list

    def read_struct(self, cmd_name):
        """Reads a set of registers defined by the memory map."""
        response = []
        for cmd, val in self.cmd_list.items():
            # Supress unused variable warning
            val = val
            if cmd.startswith(cmd_name):
                response.append(self.read_reg(cmd))
        return response

    @staticmethod
    def _parse_array(data, type_size):
        parsed_data = []
        try:
            elements = int(len(data)/type_size)
            for i in range(0, elements):
                num = int.from_bytes(data[i*type_size:(i+1)*type_size],
                                     byteorder='little')
                parsed_data.append(num)
        except (ValueError, TypeError):
            return data
        return parsed_data

    def read_reg(self, cmd_name, offset=0, size=None, to_byte_array=False):
        """Read a register defined by the memory map.

        Args:
            cmd_name(str): The name of the register to read.
            offset(int): The number of elements to offset in an array.
            size(int): The number of elements to read in an array.
        """
        cmd = self.cmd_list[cmd_name]
        response = None
        if size is None:
            if '' is cmd['total_size']:
                size = cmd['type_size']
            else:
                size = cmd['total_size']
        else:
            if '' is not cmd['total_size']:
                size = int(cmd['type_size']) * size
        if '' is not cmd['bits']:
            response = self.read_bits(cmd['offset'],
                                      cmd['bit_offset'],
                                      cmd['bits'])
        elif '' is not cmd['total_size']:
            offset = int(cmd['offset']) + (offset * int(cmd['type_size']))
            response = self.read_bytes(offset, size, True)
            response['data'] = LLMemMapIf._parse_array(response['data'],
                                                       cmd['type_size'])
        else:
            response = self.read_bytes(cmd['offset'], size, to_byte_array)
        response['msg'] = 'cmd={} response={}'.format(cmd_name,
                                                      response['msg'])
        return response

    def write_reg(self, cmd_name, data, offset=0):
        """Writes a register defined by the memory map.

        Args:
            cmd_name(str): The name of the register to read.
            data: The data to write to the register.
            offset(int): The number of elements to offset in an array.
        """
        cmd = self.cmd_list[cmd_name]
        response = None
        if '' is not cmd['bits']:
            response = self.write_bits(cmd['offset'],
                                       cmd['bit_offset'],
                                       cmd['bits'], data)
        elif '' is not cmd['total_size']:
            offset = int(cmd['offset']) + (offset * int(cmd['type_size']))
            response = self.write_bytes(offset, data)
        else:
            response = self.write_bytes(cmd['offset'], data,
                                        int(cmd['type_size']))
        response['msg'] = 'cmd={} response={}'.format(cmd_name,
                                                      response['msg'])
        return response



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Device Under Tests Shell for RIOT PAL
This module handles parsing of information from RIOT shell base tests.
"""
import logging
from .base_device import BaseDevice


class DutShell(BaseDevice):
    """Parses commands and resposes from the shell."""
    COMMAND = 'Command: '
    SUCCESS = 'Success: '
    ERROR = 'Error: '
    TIMEOUT = 'Timeout: '
    RESULT_SUCCESS = 'Success'
    RESULT_ERROR = 'Error'
    RESULT_TIMEOUT = 'Timeout'

    @staticmethod
    def _try_parse_data(data):
        if ('[' in data) and (']' in data):
            parsed_data = []
            data = data[data.find("[")+1:data.find("]")]
            data_list = data.split(', ')
            for value in data_list:
                try:
                    parsed_data.append(int(value, 0))
                except ValueError:
                    parsed_data.append(value)
            logging.debug(parsed_data)
            return parsed_data
        return None

    def send_cmd(self, send_cmd):
        """Returns a dictionary with information from the event.

        Returns:
            dict:
            The return hold dict values in the following keys::
                msg - The message from the response, only used for information.
                cmd - The command sent, used to track what has occured.
                data - Parsed information of the data requested.
                result - Either success, error or timeout.
        """
        self._write(send_cmd)
        response = self._read()
        cmd_info = {'cmd': send_cmd, 'data': None}
        while response != '':
            if self.COMMAND in response:
                cmd_info['msg'] = response.replace(self.COMMAND, '')
                cmd_info['cmd'] = cmd_info['msg'].replace('\n', '')

            if self.SUCCESS in response:
                clean_msg = response.replace(self.SUCCESS, '')
                cmd_info['msg'] = clean_msg.replace('\n', '')
                cmd_info['result'] = self.RESULT_SUCCESS
                cmd_info['data'] = self._try_parse_data(cmd_info['msg'])
                break

            if self.ERROR in response:
                clean_msg = response.replace(self.ERROR, '')
                cmd_info['msg'] = clean_msg.replace('\n', '')
                cmd_info['result'] = self.RESULT_ERROR
                break
            response = self._read()

        if response == '':
            cmd_info['result'] = self.RESULT_TIMEOUT
            logging.debug(self.RESULT_TIMEOUT)
        return cmd_info



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""packet init for RIOT PAL
This exposes usful modules in the RIOT PAL packet
"""
from pathlib import Path
from .dut_shell import DutShell
from .ll_shell import LLShell
from .ll_mem_map_if import LLMemMapIf

PHILIP_MEM_MAP_PATH = str(Path(__file__).parents[0]) + \
                      '/mem_map/philip_mem_map-00_00_01.csv'
__all__ = ['DutShell', 'LLShell', 'LLMemMapIf']



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Serial Driver for RIOT PAL
This module handles generic connection and IO to the serial driver.
"""
import os
import logging
import time
from serial import Serial, serial_for_url, SerialException


class SerialDriver:
    """Contains all reusable functions for connecting, sending and receiving
    data.  Arguments are passed through to the standard pyserial driver.  The
    defaults are changed.  Also if env variables are defined they get used as
    defaults.

    Args:
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.
    """
    DEFAULT_TIMEOUT = 1
    DEFAULT_BAUDRATE = 115200
    DEFAULT_PORT = '/dev/ttyACM0'

    def __init__(self, *args, **kwargs):
        self._connect(*args, **kwargs)
        # Used to clear the cpu and mcu buffer from startup junk data
        time.sleep(0.1)
        self.write('')
        self.read()

    def _connect(self, *args, **kwargs):
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self.DEFAULT_TIMEOUT
        if len(args) < 2:
            if 'baudrate' not in kwargs:
                if 'BAUD' in os.environ:
                    kwargs['baudrate'] = os.environ['BAUD']
                else:
                    kwargs['baudrate'] = self.DEFAULT_BAUDRATE
        if len(args) == 0:
            if 'port' not in kwargs:
                if 'PORT' in os.environ:
                    kwargs['port'] = os.environ['PORT']
                else:
                    kwargs['port'] = self.DEFAULT_PORT
        logging.debug("Serial connection args %r -- %r", args, kwargs)
        try:
            self._dev = Serial(*args, **kwargs)
        except SerialException:
            self._dev = serial_for_url(*args, **kwargs)
        self.args = args
        self.kwargs = kwargs

    def close(self):
        """Close serial connection."""
        logging.debug("Closing %s", self._dev.port)
        self._dev.close()

    def read(self):
        """Read and decode to utf-8 data.

        Returns:
            str: string of data if success, ERR string if failed.
        """
        try:
            res_bytes = self._dev.readline()
            response = res_bytes.decode("utf-8", errors="ignore")
        except (ValueError, TypeError, SerialException) as exc:
            response = 'ERR'
            logging.debug(exc)
        if response == '':
            logging.debug("Reconnecting due to timeout")
            self.close()
            self._connect(*self.args, **self.kwargs)
        logging.debug("Response: %s", response.replace('\n', ''))
        return response

    def write(self, data):
        """Writes data to a driver.  It will encode to utf-8 and add a newline

        Args:
            data(str): string or list of bytes to send to the driver.
        """
        # Clear the input buffer in case it junk data go in creating an offset
        self._dev.reset_input_buffer()
        logging.debug("Sending: " + data)
        self._dev.write((data + '\n').encode('utf-8'))
